// written by Michael Buro (2003)
// license: GPL

#include "MinMaxLP.H"
#include "SimplexLongDouble.H"
#include "SimplexMPQ.H"
#include <iostream>
#include <gmpxx.h>

using namespace std;

typedef long double NUMBER;
typedef mpq_class   NUMBER2;

// denominator of rational approximation at most:
#define MAX_DENOM "10000000000" // 8


int minmaxlp(int n1, int n2, const std::vector<double> &input,
	     std::vector<double> &output, double &max_val)
{
  int res = 0;
  
  if (n1*n2 != int(input.size())) { cerr << "size != n*m" << endl; return 2; }

  // construct matrix A  NUMBER **A;

  NUMBER **A = new NUMBER*[n1];
  for (int i=0; i < n1; ++i) A[i] = new NUMBER[n2];

  double mat_min = 0, mat_max = 0, adj = 0;
  int i = 0;
  for (int m1=0; m1 < n1; ++m1) {
    for (int m2=0; m2 < n2; ++m2) {
      mat_min = ::min(mat_min, input[i]);
      mat_max = ::max(mat_max, input[i]);      
      A[m1][m2] = input[i];
      ++i;
    }
  }

  //cout << mat_min << " " << mat_max << endl;
  
  // adjust all matrix value to be >= 0.5
  
  if (mat_min < 0.5) {

    adj = - mat_min + 1.0;

    for (int m1=0; m1 < n1; ++m1) {
      for (int m2=0; m2 < n2; ++m2) {
	A[m1][m2] += adj;
      }
    }

    mat_min += adj;
    mat_max += adj;

    // cout << "a: " << adj << " " << mat_min << " " << mat_max << endl;
  }
  
  /*

  linear program for first player
            
  max Z
  Z - sum_i(0..n1-1) A(i,j) * x_i <= 0  for all j
  sum_i(0..n1-1) x_i = 1
  x_i >= 0

  */

  {
    // generate simplex input

    const int n  = n1+1;  // #variables (probabilities for moves 0..n1-1, Z)
    const int m1 = n2;    // # <= constraints
    const int m2 = 0;     // # >= constraints
    const int m3 = 1;     // # == constraints
    const int m = m1+m2+m3;

    // B ~ (m+1)x(n+1) matrix, first column: bounds
              
    vector< vector<NUMBER> > B(m+1); // objective function + constraints

    // B=0 (important: the code below assumes 0 values!)
    
    NUMBER x0 = 0;
    for (int i=0; i < m+1; ++i) {
      B[i].reserve(n+1);
      for (int j=0; j < n+1; ++j) B[i].push_back(x0);
    }

    // 1. objective function

    B[0][1] = 1;  // Z
              
    // 2. <= constraints

    for (int j=0; j < n2; j++) {
      B[j+1][0] = 1;  // <= 1
      B[j+1][1] = 1;  // Z
      for (int i=0; i < n1; i++) B[j+1][i+2] = - A[i][j];
    }

    // 3. == constraint

    B[m][0] = 1;
    for (int i=0; i < n1; ++i) B[m][i+2] = 1;
              
    NUMBER opt_val;
    vector<NUMBER> result;

    // cerr << "solve: " << n1 << " " << n2 << endl;
    res = SimplexLongDouble::solve(B, m1, m2, m3, opt_val, result);
    opt_val--;

    bool dubious = false;
              
    if (!res) {
                
      // check

      if (opt_val > mat_max+0.0001 || opt_val < mat_min-0.0001) {
	cerr << "val= " << opt_val << endl;
	dubious = true; goto check;
      }
                
      NUMBER s = 0;

      cerr.precision(20);
                
      for (unsigned int i=1; i < result.size(); ++i) {
	if (result[i] < -0.00001 || result[i] > 1.00001) {
	  cerr << "p= " << result[i] << endl;
	  dubious = true; goto check;
	}
	s += result[i];
      }

      s -= 1;
                
      if (!dubious && (s > 1e-6 || s < -1e-6)) {
	cerr << "s= " << (s+1) << endl;
	dubious = true;
      }
    }

  check:;
              
    if (res != 0 || dubious) {

      // if double lp solve failed or the result
      // is dubious switch to rational arithmetic
                
      if (dubious)
	cerr << "dubious result" << endl;                 
      else
	cerr << "simplex failed " << res << endl;

      // try again with more precision
                
      vector< vector<NUMBER2> > B2(m+1);

      NUMBER2 x0 = 0;
      for (int i=0; i < m+1; ++i) {
	B2[i].reserve(n+1);
	for (int j=0; j < n+1; ++j) {
	  B2[i].push_back(x0);
	}
      }

      for (int i=0; i < m+1; ++i) {
	for (int j=0; j < n+1; ++j)
	  B2[i][j] = (double)B[i][j];
      }
      NUMBER2 opt_val2;
      vector<NUMBER2> result2;
                
      int res = SimplexMPQ::solve(B2, m1, m2, m3, opt_val2, result2);
      cerr << "A" << endl;
      opt_val2--;

      if (res != 0) {
	cerr << "simplex2 failed - giving up " << res << endl;
	goto clean_up;
      }
                
      opt_val = opt_val2.get_d();

      if (opt_val < -1 || opt_val > 1) {
	cerr << "opt_val= " << opt_val << " corrected" << endl;
	if      (opt_val < -1) opt_val = -1;
	else if (opt_val > +1) opt_val = +1;
      }
                
      result.clear();
                
      for (unsigned int i=0; i < result2.size(); ++i) {
	double r = result2[i].get_d();

	if (i > 0) {
	  if (r < 0 || r > 1) {
	    cerr << "p= " << r << " corrected" << endl;
	    if      (r < 0) r = 0;
	    else if (r > 1) r = 1;
	  }
	}
                  
	result.push_back(r);
      }
      cerr << "B" << endl;              
    }

    // check: sum of probabilities should be close to 1

    NUMBER s = 0;

    for (unsigned int i=1; i < result.size(); ++i) s += result[i];
    s -= 1;
    if (s < -1e-5 || s > 1e-5) {
      cerr << "****** s= " << s << endl;
    }

    // adjust values: probs in [0,1], val in [-1,1]
    // snap probs. to 0 and 1
    // snap value to 0
              
    const double eps = 1e-7;
    for (unsigned int i=1; i < result.size(); ++i) {
      if      (result[i] < 0)       result[i] = 0;
      else if (result[i] > 1)       result[i] = 1;
      else if (result[i] < eps)     result[i] = 0;
      else if (result[i] > 1.0-eps) result[i] = 1;
    }


    // clip value to [mat_min, mat_max]
    
    bool fixed = false;
    if      (opt_val < mat_min) { fixed = true; opt_val = mat_min; }
    else if (opt_val > mat_max) { fixed = true; opt_val = +mat_max; }
    else if (opt_val < eps && opt_val > -eps) opt_val = 0;

    if (fixed) cout << "MinMaxLP: val fixed" << endl;
    
#if MP_TYPE
    {
      // round opt_val
      // denominator <= N!

      mpz_class N(MAX_DENOM);                           // 8

      if (opt_val.get_den() > N) {

	//cout << "before " << opt_val.get_d() << endl;

	opt_val.get_num() = (opt_val.get_num()*N)/opt_val.get_den();
	opt_val.get_den() = N;

	//cout << "after " << opt_val.get_d() << endl;            
      }

      opt_val.canonicalize();
    }
#endif
              
#if TEST
    cerr << " val: " << opt_val << endl;
#endif

    // cout << "adj: " << adj << endl;
    
    max_val = opt_val - adj;

    output.clear();
    for (unsigned int i=1; i < result.size(); ++i) output.push_back(result[i]);
  }

#if 0
  
  // no need for solving lp for second player!
            
  /*

  linear program for second player
            
  min Z
  Z - sum_j(0..n2-1) A(i,j) * y_j >= 0  for all i
  sum_j(0..n2-1) y_j = 1
  y_j >= 0

  */

#endif

  res = 0;
  
 clean_up:;
  
  for (int i=0; i < n1; ++i) delete [] A[i];
  delete [] A;
  return res;
}
