// Simplex LP solver
// based on Numerical Recipes code

// adjusted by Michael Buro (2003)

#include <math.h>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include "Simplex.H"

using namespace std;

#define NR_END 1
#define FREE_ARG char*
#define FREEALL free_ivector(l3,1,m); free_ivector(l1,1,n+1);

static void my_abs(SIMPLEX_NUMBER &x) { if (x < 0) x = -x; }

static void nrerror(char error_text[])
{ 
  fprintf(stderr,"Numerical Recipes run-time error...\n");
  fprintf(stderr,"%s\n",error_text);
  fprintf(stderr,"...now exiting to system...\n");
  exit(1);
}

static int *ivector(long nl, long nh)
/* allocate an int vector with subscript range v[nl..nh] */
{
  int *v;

  v= (int *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(int)));
  if (!v) nrerror("allocation failure in ivector()");
  return v-nl+NR_END;
}

static void free_ivector(int *v, long nl, long nh)
/* free an int vector allocated with ivector() */
{
  free((FREE_ARG) (v+nl-NR_END));
}

static void simp1(SIMPLEX_NUMBER **a, int mm, int ll[], int nll, int iabf, int *kp, SIMPLEX_NUMBER *bmax)
{
  int k;
  SIMPLEX_NUMBER test;

  if (nll <= 0) *bmax = 0;
  else {
  
    *kp = ll[1];
    *bmax = a[mm+1][*kp+1];

    for (k=2; k <= nll; k++) {

      if (iabf == 0)
	test = a[mm+1][ll[k]+1] - (*bmax);
      else {
	SIMPLEX_NUMBER x = a[mm+1][ll[k]+1];
	SIMPLEX_NUMBER y = *bmax;
	my_abs(x); my_abs(y);
	// was: test = fabs(a[mm+1][ll[k]+1]) - fabs(*bmax);
	test = x - y;
      }
    
      if (test > 0) {
	*bmax = a[mm+1][ll[k]+1];
	*kp = ll[k];
      }
    }
  }
}

static void simp2(SIMPLEX_NUMBER **a, int m, int n, int *ip, int kp)
{
  int k, i;
  SIMPLEX_NUMBER qp=0, q0=0, q, q1;

  *ip = 0;

  for (i=1; i <= m; i++) {

    //cout << "a " << (i+1) << " " << (kp+1) << " = " << a[i+1][kp+1] << endl;

    if (a[i+1][kp+1] < -SIMPLEX_EPS) break;
  }
  // cout << "-> " << i << " " << m << endl;
  
  if (i > m) return;

  q1 = - a[i+1][1] / a[i+1][kp+1];
  *ip = i;

  for (i=*ip+1; i <= m; i++) {

    if (a[i+1][kp+1] < -SIMPLEX_EPS) {

      q = - a[i+1][1] / a[i+1][kp+1];

      if (q < q1) {

	*ip = i;
	q1 = q;

      } else if (q == q1) {

	for (k=1; k <= n; k++) {
	  qp = - a[*ip+1][k+1] / a[*ip+1][kp+1];
	  q0 = - a[i+1][k+1] / a[i+1][kp+1];
	  if (q0 != qp) break;
	}

	if (q0 < qp) *ip = i;
      }
    }
  }
}

static void simp3(SIMPLEX_NUMBER **a, int i1, int k1, int ip, int kp)
{
  int kk, ii;
  SIMPLEX_NUMBER piv;

  piv = 1.0/a[ip+1][kp+1];
  
  for (ii=1; ii <= i1+1; ii++)
    if (ii-1 != ip) {
      a[ii][kp+1] *= piv;
      for (kk=1; kk <= k1+1; kk++)
	if (kk-1 != kp) a[ii][kk] -= a[ip+1][kk] * a[ii][kp+1];
    }
  
  for (kk=1; kk <= k1+1; kk++) if (kk-1 != kp) a[ip+1][kk] *= -piv;
  a[ip+1][kp+1]=piv;
}


/* weird input format:

   objective function    (0, normal x_i weights)
   constraints           (bound, inverted x_i weights!)

 */

static void simplx(SIMPLEX_NUMBER **a, int m, int n, int m1, int m2, int m3, int *icase,
		   int izrov[], int iposv[])
{
  int i, ip, is, k, kh, kp, nl1;
  int *l1, *l3;
  SIMPLEX_NUMBER q1, bmax;

#if 0

  cout << "simplx matrix" << endl;
  for (int j=1; j <= m+2; j++) {
    for (int i=1; i <= n+1; i++) {
      cout << a[j][i] << " ";
    }
    cout <<endl;
  }

#endif
  
  if (m != m1+m2+m3) nrerror("Bad input constraint counts in simplex");
  l1=ivector(1,n+1);
  l3=ivector(1,m);

  nl1 = n;

  for (k=1; k <= n; k++) l1[k] = izrov[k] = k;

  for (i=1; i <= m; i++) {
    if (a[i+1][1] < 0) nrerror("Bad input tableau in simplex");
    iposv[i] = n+i;
  }
  
  for (i=1; i <= m2; i++) l3[i] = 1;

  if (m2+m3) {
    
    for (k=1; k <= n+1; k++) {
      q1 = 0;
      for (i=m1+1; i <= m; i++) q1 += a[i+1][k];
      a[m+2][k] = -q1;
    }
    
    for (;;) {

      simp1(a, m+1, l1, nl1, 0, &kp, &bmax);
      
      if (bmax <= SIMPLEX_EPS && a[m+2][1] < -SIMPLEX_EPS) {

	*icase = -1;
	FREEALL return;

      } else if (bmax <= SIMPLEX_EPS && a[m+2][1] <= SIMPLEX_EPS) {

	for (ip=m1+m2+1; ip <= m; ip++) {
	  if (iposv[ip] == ip+n) {
	    simp1(a, ip, l1, nl1, 1, &kp, &bmax);
	    if (bmax > SIMPLEX_EPS) goto one;
	  }
	}

	for (i=m1+1; i <= m1+m2; i++)
	  if (l3[i-m1] == 1)
	    for (k=1; k <= n+1; k++) a[i+1][k] = -a[i+1][k];

	break;
      }
      
      simp2(a, m, n, &ip, kp);

      if (ip == 0) {
	*icase = -1;
	FREEALL return;
      }
      
    one: ;
      
      simp3(a, m+1, n, ip, kp);

      if (iposv[ip] >= n+m1+m2+1) {

	for (k=1; k <= nl1; k++) if (l1[k] == kp) break;

	--nl1;

	for (is=k; is <= nl1; is++) l1[is] = l1[is+1];
	
      } else {

	kh = iposv[ip]-m1-n;
	if (kh >= 1 && l3[kh]) {
	  l3[kh] = 0;
	  ++a[m+2][kp+1];
	  for (i=1;i <= m+2; i++) a[i][kp+1] = -a[i][kp+1];
	}
      }
      is=izrov[kp];
      izrov[kp] = iposv[ip];
      iposv[ip] = is;
      
    }
  }

  time_t start_time = time(0);
  
  for (;;) {

    time_t now = time(0);

    if (now - start_time > SIMPLEX_MAX_TIME) {
      *icase = 2;
      goto end;
    }
    
    simp1(a, 0, l1, nl1, 0, &kp, &bmax);

    //cerr << kp << " " << flush;

    if (bmax <= 0) {
      *icase = 0;
      goto end;
    }

    simp2(a, m, n, &ip, kp);

    if (ip == 0) {
      *icase = 1;
      goto end;
    }

    simp3(a, m, n, ip, kp);

    is = izrov[kp];
    izrov[kp] = iposv[ip];
    iposv[ip] = is;
  }

 end:;

  //cerr << endl;
  FREEALL;
  return;
}


int SIMPLEX_TYPE::solve(vector< vector<SIMPLEX_NUMBER> > &A, int m1, int m2, int m3,
		    SIMPLEX_NUMBER &opt_val, vector<SIMPLEX_NUMBER> &results)
{
  // create (M+2) x (N+1) matrix

  int m = m1+m2+m3;
  int rows = A.size();
  if (rows != m+1) {
    cerr << "tableau corrupt (rows)" << endl; exit(-1);
  }

  int columns = A[0].size();
  int n = columns - 1;  // #variables
  
  for (int i=1; i < rows; i++) {
    if (static_cast<int>(A[i].size()) != columns) {
      cerr << "tableau corrupt (columns)" << endl; exit(-1);
    }
  }

  SIMPLEX_NUMBER **a;

  a = new SIMPLEX_NUMBER*[rows+2]; // 2 = scratch space + min.index = 1

  for (int i=0; i < rows+2; i++) {  // 1 = min. index = 1
    a[i] = new SIMPLEX_NUMBER[columns+1];
  }
    
  for (int j=0; j < rows; j++)
    for (int i=0; i < columns; i++) {
      if (i > 0 && j > 0)
	a[j+1][i+1] = - A[j][i];
      else
	a[j+1][i+1] = A[j][i];
    }

  int icase;
  int *izrov = ivector(0, n);
  int *iposv = ivector(0, m);

  simplx(a, m, n, m1, m2, m3, &icase, izrov, iposv);

#if 0  
  cout << "rows:" << (rows+2) << " columns: " << (columns+1) << " " << icase << endl;
  for (int i=1; i <= n; i++) cout << izrov[i] << " ";
  cout << endl;
  for (int i=1; i <= m; i++) cout << iposv[i] << " ";
  cout << endl;

  cout << "matrix:" << endl;
  
  for (int j=0; j < rows+2; j++) {
    for (int i=0; i < columns+1; i++) {
      cout << a[j][i] << " ";
    }
    cout << endl;
  }
#endif
  
  results.clear();
  results.reserve(n);
  for (int i=0; i < n; i++) {
    SIMPLEX_NUMBER x = -1;
    results.push_back(x);
  }

  for (int i=1; i <= m; i++) {
    if (iposv[i] > n) continue;
    results[iposv[i]-1] = a[i+1][1];
  }

  for (int i=0; i < n; i++) {
    if (results[i] < 0) results[i] = 0;
  }

  opt_val = a[1][1];

  // clean up
  
  free_ivector(iposv, 0, m);
  free_ivector(izrov, 0, n);

  for (int i=0; i < rows+2; i++) delete[] a[i];
  delete[] a;
 
  return icase;
}


#if 0

// test cases

#if 0

double v[5][5] = {
  {   0,    1,  1,  3, -0.5 },
  { 740,    1,  0,  2,  0   },
  {   0,    0,  2,  0,  -7   },
  {   0.5,  0,  1,  -1, 2   },
  {   9,    1,  1,  1,  1   }
};

const int n  = 4;   // variables
const int m1 = 2;   // <=
const int m2 = 1;   // >=
const int m3 = 1;   // ==

#else
#if 0

// cycle?

double v[4][5] = {
  {   0,    10, -57, -9, -24 },
  {   0,   0.5, -5.5, -2.5,  9 },
  {   0,   0.5, -1.5, -0.5,  1 },
  {   1,  1, 0,  0, 0   },
};

const int n  = 4;   // variables
const int m1 = 3;   // <=
const int m2 = 0;   // >=
const int m3 = 0;   // ==

#else
#if 0

double v[3][2] = {
  {   0,    1,  }, 
  {   1,    1,  },
};

const int n  = 1;   // variables
const int m1 = 1;   // <=
const int m2 = 0;   // >=
const int m3 = 0;   // ==

#else

double v[4][4] = {
  {   0,  -1, 1, 2 },
  {  23,   3, 4, -3  },
  {  10,   5, -4, -3 },
  {  30,  7, 4, 11 },
};

const int n  = 3;   // variables
const int m1 = 3;   // <=
const int m2 = 0;   // >=
const int m3 = 0;   // ==

#endif
#endif
#endif


int main()
{
  const int m = m1+m2+m3;

  vector< vector<SIMPLEX_NUMBER> > A(m+1);

  for (int i=0; i < m+1; i++) {
    A[i].reserve(n+1);
    for (int j=0; j < n+1; j++) {
      SIMPLEX_NUMBER x = v[i][j];
      A[i].push_back(x);
    }
  }

  SIMPLEX_NUMBER opt_val;
  vector<SIMPLEX_NUMBER> result;
  
  int res = Simplex::solve(A, m1, m2, m3, opt_val, result);

  cout << "res " << res << endl;
  cout << "opt_val " << opt_val << endl;
  for (int i=0; i < n; i++) {
    cout << result[i] << " ";
  }   
  cout << endl;
  
  return 0;
}

#endif

